# vim:ts=8:sts=4:sw=4
# -*- coding: utf-8 -*-

import gobject
import socket
import struct
import random

import drawing

(
    CMD_CONNECTED,
    CMD_READY,
    CMD_START,
    CMD_END,
    CMD_STOP,
    CMD_START_TURN,
    CMD_REGISTER,
    CMD_REGISTER_FAIL,
    CMD_TRY_FINISH_TURN,
    CMD_FINISH_TURN,
    CMD_FINISH_TURN_FAIL,
    CMD_GIVE_1_TILE,
    CMD_GIVE_3_TILE,
    CMD_MOVE_DECK_TO_BOARD,
    CMD_MOVE_BOARD_TO_DECK,
    CMD_MOVE_BOARD_TO_BOARD,
    CMD_EXCHANGE_DECK_TO_BOARD,
    CMD_EXCHANGE_BOARD_TO_DECK,
    CMD_EXCHANGE_BOARD_TO_BOARD,
    CMD_MOVE_FAIL,
    CMD_CHAT,
    CMD_QUIT
) = range(22)

(
    SERVER_CREATED,
    SERVER_READY,
    SERVER_STARTED
) = range(3)

(
    CLIENT_CONNECTED,
    CLIENT_READY,
    CLIENT_STARTED,
    CLIENT_REGISTERED
) = range(4)

def send_cmd (sock, cmd, *args):
    if cmd == CMD_CONNECTED:
	id, = args
	argfmt = 'h'
	argbuf = struct.pack(argfmt, id)
    elif cmd == CMD_READY:
	id, name = args
	namelen = len(name)
	argfmt = '2h' + str(namelen) + 's'
	argbuf = struct.pack(argfmt, id, namelen, name)
    elif cmd == CMD_START:
	tiles, = args
	argfmt = '14h'
	argbuf = ''
	for i in tiles:
	    argbuf = argbuf + struct.pack('h', i)
    elif cmd == CMD_END:
	score, totalscore, remain = args
	argbuf = struct.pack('h', len(score))
	argnum = 1
	for id in score.keys():
	    argbuf += struct.pack('h', id)
	    argnum += 1
	for sc in score.values():
	    argbuf += struct.pack('h', sc)
	    argnum += 1
	for tot in totalscore.values():
	    argbuf += struct.pack('h', tot)
	    argnum += 1
	for tiles in remain.values():
	    argbuf += struct.pack('h', len(tiles))
	    argnum += 1
	    for t in tiles:
		argbuf += struct.pack('h', t)
		argnum += 1
	argfmt = str(argnum) + 'h'
    elif cmd == CMD_STOP:
	id, = args
	argfmt = 'h'
	argbuf = struct.pack(argfmt, id)
    elif cmd == CMD_START_TURN:
	id, = args
	argfmt = 'h'
	argbuf = struct.pack(argfmt, id)
    elif cmd == CMD_REGISTER:
	id, = args
	argfmt = 'h'
	argbuf = struct.pack(argfmt, id)
    elif cmd == CMD_REGISTER_FAIL:
	id, = args
	argfmt = 'h'
	argbuf = struct.pack(argfmt, id)
    elif cmd == CMD_TRY_FINISH_TURN:
	id, = args
	argfmt = 'h'
	argbuf = struct.pack(argfmt, id)
    elif cmd == CMD_FINISH_TURN:
	id, type = args
	argfmt = '2h'
	argbuf = struct.pack(argfmt, id, type)
    elif cmd == CMD_FINISH_TURN_FAIL:
	id, = args
	argfmt = 'h'
	argbuf = struct.pack(argfmt, id)
    elif cmd == CMD_GIVE_1_TILE:
	id, tile = args
	argfmt = '2h'
	argbuf = struct.pack(argfmt, id, tile)
    elif cmd == CMD_GIVE_3_TILE:
	id, tile1, tile2, tile3 = args
	argfmt = '4h'
	argbuf = struct.pack(argfmt, id, tile1, tile2, tile3)
    elif cmd == CMD_MOVE_DECK_TO_BOARD:
	id, tile, row, col = args
	argfmt = '4h'
	argbuf = struct.pack(argfmt, id, tile, row, col)
    elif cmd == CMD_MOVE_BOARD_TO_DECK:
	id, tile, row, col = args
	argfmt = '4h'
	argbuf = struct.pack(argfmt, id, tile, row, col)
    elif cmd == CMD_MOVE_BOARD_TO_BOARD:
	id, tile, frow, fcol, row, col = args
	argfmt = '6h'
	argbuf = struct.pack(argfmt, id, tile, frow, fcol, row, col)
    elif cmd == CMD_EXCHANGE_DECK_TO_BOARD:
	id, ftile, ttile, row, col = args
	argfmt = '5h'
	argbuf = struct.pack(argfmt, id, ftile, ttile, row, col)
    elif cmd == CMD_EXCHANGE_BOARD_TO_DECK:
	id, ftile, ttile, row, col = args
	argfmt = '5h'
	argbuf = struct.pack(argfmt, id, ftile, ttile, row, col)
    elif cmd == CMD_EXCHANGE_BOARD_TO_BOARD:
	id, ftile, ttile, frow, fcol, row, col = args
	argfmt = '7h'
	argbuf = struct.pack(argfmt, id, ftile, ttile, frow, fcol, row, col)
    elif cmd == CMD_MOVE_FAIL:
	id, = args
	argfmt = 'h'
	argbuf = struct.pack(argfmt, id)
    elif cmd == CMD_CHAT:
	id, chat = args
	chatlen = len(chat)
	argfmt = '2h' + str(chatlen) + 's'
	argbuf = struct.pack(argfmt, id, chatlen, chat)
    elif cmd == CMD_QUIT:
	id, = args
	argfmt = 'h'
	argbuf = struct.pack(argfmt, id)

    fmt = '2h'
    argsize = struct.calcsize(argfmt)
    buf = struct.pack(fmt, cmd, argsize)
    sock.send(buf + argbuf)

def recv_cmd (sock):
    size = struct.calcsize('2h')
    buf = sock.recv(size)
    if not buf:
	raise Disconnect

    cmd, argsize = struct.unpack('2h', buf)
    argbuf = sock.recv(argsize)

    if cmd == CMD_CONNECTED:
	args = struct.unpack('h', argbuf)
    elif cmd == CMD_READY:
	size = struct.calcsize('2h')
	id, namelen = struct.unpack('2h', argbuf[:size])
	name, = struct.unpack(str(namelen) + 's', argbuf[size:])
	args = id, name
    elif cmd == CMD_START:
	args = list(struct.unpack('14h', argbuf)),
    elif cmd == CMD_END:
	remain = {}
	size = struct.calcsize('h')
	buf = argbuf
	idnum, = struct.unpack('h', buf[:size])
	buf = buf[size:]
	fmt = str(idnum) + 'h'
	ids = list(struct.unpack(fmt, buf[:size * idnum]))
	buf = buf[size * idnum:]
	scs = list(struct.unpack(fmt, buf[:size * idnum]))
	buf = buf[size * idnum:]
	score = dict([(id, sc) for id, sc in zip(ids, scs)])
	tots = list(struct.unpack(fmt, buf[:size * idnum]))
	buf = buf[size * idnum:]
	totalscore = dict([(id, tot) for id, tot in zip(ids, tots)])
	for id in ids:
	    tilenum, = struct.unpack('h', buf[:size])
	    buf = buf[size:]
	    fmt = str(tilenum) + 'h'
	    remain[id] = list(struct.unpack(fmt, buf[:size * tilenum]))
	    buf = buf[size * tilenum:]
	args = score, totalscore, remain
    elif cmd == CMD_STOP:
	args = struct.unpack('h', argbuf)
    elif cmd == CMD_START_TURN:
	args = struct.unpack('h', argbuf)
    elif cmd == CMD_REGISTER:
	args = struct.unpack('h', argbuf)
    elif cmd == CMD_REGISTER_FAIL:
	args = struct.unpack('h', argbuf)
    elif cmd == CMD_TRY_FINISH_TURN:
	args = struct.unpack('h', argbuf)
    elif cmd == CMD_FINISH_TURN:
	args = struct.unpack('2h', argbuf)
    elif cmd == CMD_FINISH_TURN_FAIL:
	args = struct.unpack('h', argbuf)
    elif cmd == CMD_GIVE_1_TILE:
	args = struct.unpack('2h', argbuf)
    elif cmd == CMD_GIVE_3_TILE:
	args = struct.unpack('4h', argbuf)
    elif cmd == CMD_MOVE_DECK_TO_BOARD:
	args = struct.unpack('4h', argbuf)
    elif cmd == CMD_MOVE_BOARD_TO_DECK:
	args = struct.unpack('4h', argbuf)
    elif cmd == CMD_MOVE_BOARD_TO_BOARD:
	args = struct.unpack('6h', argbuf)
    elif cmd == CMD_EXCHANGE_DECK_TO_BOARD:
	args = struct.unpack('5h', argbuf)
    elif cmd == CMD_EXCHANGE_BOARD_TO_DECK:
	args = struct.unpack('5h', argbuf)
    elif cmd == CMD_EXCHANGE_BOARD_TO_BOARD:
	args = struct.unpack('7h', argbuf)
    elif cmd == CMD_MOVE_FAIL:
	args = struct.unpack('h', argbuf)
    elif cmd == CMD_CHAT:
	size = struct.calcsize('2h')
	id, chatlen = struct.unpack('2h', argbuf[:size])
	chat, = struct.unpack(str(chatlen) + 's', argbuf[size:])
	args = id, chat
    elif cmd == CMD_QUIT:
	args = struct.unpack('h', argbuf)

    return cmd, args

class Gamer:
    def __init__ (self, id):
	self.id = id
	self.tiles = []

class Disconnect (Exception):
    pass

class Server:
    gamers = {}
    veteran = {}
    sequence = 0
    number = 0
    timeoutid = None

    copyboard = []
    copydeck = []

    def __init__ (self, xml, drawing):
	self.xml = xml
	self.xml.signal_autoconnect(self)
	self.drawing = drawing

    def run (self, port):
	self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	try:
	    self.sock.bind(('', port))
	except:
	    self.sock.close()
	    raise
	    
	self.sock.listen(1)
	self.io = gobject.io_add_watch(self.sock,
		gobject.IO_IN | gobject.IO_ERR | gobject.IO_HUP, self.accept)
	self.setup()

    def accept (self, sock, condition):
	connsock, addr = sock.accept()
	io = gobject.io_add_watch(connsock,
		gobject.IO_IN | gobject.IO_ERR | gobject.IO_HUP, self.receive)

	self.connect(connsock, io)
	return True
    
    def receive (self, sock, condition):
	try:
	    cmd, args = recv_cmd(sock)
	except Disconnect:
	    self.disconnect(sock)
	    return False

	if cmd == CMD_READY:
	    id, name = args
	    self.recv_ready(sock, id, name)
	elif cmd == CMD_REGISTER:
	    id, = args
	    self.recv_register(sock, id)
	elif cmd == CMD_TRY_FINISH_TURN:
	    id, = args
	    self.recv_try_finish_turn(sock, id)
	elif cmd == CMD_MOVE_DECK_TO_BOARD:
	    id, tile, row, col = args
	    self.recv_move_deck_to_board(sock, id, tile, row, col)
	elif cmd == CMD_MOVE_BOARD_TO_DECK:
	    id, tile, row, col = args
	    self.recv_move_board_to_deck(sock, id, tile, row, col)
	elif cmd == CMD_MOVE_BOARD_TO_BOARD:
	    id, tile, frow, fcol, row, col = args
	    self.recv_move_board_to_board(sock, id, tile, frow, fcol, row, col)
	elif cmd == CMD_EXCHANGE_DECK_TO_BOARD:
	    id, ftile, ttile, row, col = args
	    self.recv_exchange_deck_to_board(sock, id, ftile, ttile, row, col)
	elif cmd == CMD_EXCHANGE_BOARD_TO_DECK:
	    id, ftile, ttile, row, col = args
	    self.recv_exchange_board_to_deck(sock, id, ftile, ttile, row, col)
	elif cmd == CMD_EXCHANGE_BOARD_TO_BOARD:
	    id, ftile, ttile, frow, fcol, row, col = args
	    self.recv_exchange_board_to_board(sock, id,
		    ftile, ttile, frow, fcol, row, col)
	elif cmd == CMD_CHAT:
	    id, chat = args
	    self.recv_chat(sock, id, chat)

	return True

    def connect (self, sock, io):
	id = self.sequence
	self.sequence = self.sequence + 1

	self.gamers[id] = Gamer(id)
	self.gamers[id].state = CLIENT_CONNECTED
	self.gamers[id].io = io
	self.gamers[id].sock = sock
	self.gamers[id].score = []
	self.gamers[id].totalscore = 0

	self.created()
	send_cmd(sock, CMD_CONNECTED, id)
    
    def setup (self):
	self.state = SERVER_CREATED
	self.turn = []
	self.round = 0
	if self.timeoutid:
	    gobject.source_remove(self.timeoutid)
	    self.timeoutid = None
	self.register_turn = False
	self.drawing.show_control()

    def created (self):
	self.state = SERVER_CREATED
	self.drawing.set_start_game(False)
	self.drawing.set_end_game(False)

    def ready (self):
	self.state = SERVER_READY
	self.drawing.set_start_game(True)
	self.drawing.set_end_game(False)

    def start (self):
	self.state = SERVER_STARTED
	self.drawing.set_start_game(False)
	self.drawing.set_end_game(True)
	self.tiles = []
	for i in range(self.drawing.board.ROWS):
	    r = []
	    for j in range(self.drawing.board.COLS):
		r.append(-1)
	    self.tiles.append(r)

	self.tilestack = drawing.get_tile_list()
	random.shuffle(self.tilestack)
	for i, g in self.gamers.iteritems():
	    if g.state != CLIENT_READY:
		continue
	    g.state = CLIENT_STARTED
	    g.tiles = []
	    for n in range(14):
		g.tiles.append(self.tilestack.pop(0))
	    send_cmd(g.sock, CMD_START, g.tiles)
	self.start_turn()

    def end (self):
	self.ready()

	self.turn = []
	if self.timeoutid:
	    gobject.source_remove(self.timeoutid)
	    self.timeoutid = None
	self.register_turn = False

	remain = {}
	score = {}
	for i, g in self.gamers.iteritems():
	    if g.state == CLIENT_CONNECTED:
		continue
	    if g.state == CLIENT_READY:
		continue
	    while len(g.score) <= self.round:
		g.score.append(0)

	    for t in g.tiles:
		if t == -1:
		    continue
		color, number = drawing.get_tile_tuple(t)
		if color == 4:
		    g.score[self.round] -= 30
		else:
		    g.score[self.round] -= number + 1

	    remain[i] = g.tiles
	    score[i] = g.score[self.round]
	minscore = max(score.values())
	sumscore = sum(score.values())
	mincount = score.values().count(minscore)

	totalscore = {}
	for i, g in self.gamers.iteritems():
	    if g.state == CLIENT_CONNECTED:
		continue
	    if g.state == CLIENT_READY:
		continue
	    if score[i] == minscore:
		score[i] = -(sumscore - minscore * mincount) / mincount
		g.score[self.round] = score[i]

	    g.totalscore += g.score[self.round]
	    totalscore[i] = g.totalscore

	newscore = max(totalscore.values())
	for i, g in self.gamers.iteritems():
	    if g.state == CLIENT_CONNECTED:
		continue
	    if g.state == CLIENT_READY:
		g.totalscore = newscore
	    else:
		g.state = CLIENT_READY
	    send_cmd(g.sock, CMD_END, score, totalscore, remain)
	self.round += 1

    def stop (self):
	if self.number >= 2:
	    self.ready()
	else:
	    self.created()

	self.turn = []
	if self.timeoutid:
	    gobject.source_remove(self.timeoutid)
	    self.timeoutid = None
	self.register_turn = False

	for i, g in self.gamers.iteritems():
	    if g.state == CLIENT_CONNECTED:
		continue
	    g.state = CLIENT_READY
	    send_cmd(g.sock, CMD_STOP, i)

	self.round += 1

    def quit (self, id):
	gamer = self.gamers[id]
	del self.gamers[id]

	if gamer.state == CLIENT_CONNECTED:
	    return

	self.number = self.number - 1
	if self.state == SERVER_READY:
	    if self.number < 2:
		self.created()
	if self.state == SERVER_STARTED:
	    if self.number < 2:
		self.stop()
	    else:
		self.tilestack.extend(gamer.tiles)

		if self.turn[0] == id:
		    self.turn.pop(0)
		    if self.timeoutid:
			gobject.source_remove(self.timeoutid)
			self.timeoutid = None
		    self.start_turn()

		elif id in self.turn:
		    self.turn.remove(id)

	for i, g in self.gamers.iteritems():
	    if g.state == CLIENT_CONNECTED:
		continue
	    send_cmd(g.sock, CMD_QUIT, id)
	self.veteran[id] = gamer
	if self.number == 0:
	    gobject.source_remove(self.io)
	    self.sock.close()

    def start_turn (self):
	if not self.turn:
	    for i, g in self.gamers.iteritems():
		if g.state != CLIENT_STARTED and g.state != CLIENT_REGISTERED:
		    continue
		self.turn.append(i)
	id = self.turn[0]

	self.copyboard = []
	for r in self.tiles[:]:
	    self.copyboard.append(r[:])
	self.copydeck = self.gamers[id].tiles[:]

	for i, g in self.gamers.iteritems():
	    if g.state != CLIENT_STARTED and g.state != CLIENT_REGISTERED:
		continue
	    send_cmd(g.sock, CMD_START_TURN, id)
	self.timeoutid = gobject.timeout_add(60000, self.timeout)

    def timeout (self):
	self.timeoutid = None

	id = self.turn[0]
	sock = self.gamers[id].sock
	self.finish_turn(sock, id, 2)
	return False

    def register (self, sock, id):
	self.gamers[id].state = CLIENT_REGISTERED
	self.register_turn = True

	self.copyboard = []
	for r in self.tiles[:]:
	    self.copyboard.append(r[:])
	self.copydeck = self.gamers[id].tiles[:]

	for i, g in self.gamers.iteritems():
	    if g.state != CLIENT_STARTED and g.state != CLIENT_REGISTERED:
		continue
	    send_cmd(g.sock, CMD_REGISTER, id)

    def register_fail (self, sock, id):
	for i, g in self.gamers.iteritems():
	    if g.state != CLIENT_STARTED and g.state != CLIENT_REGISTERED:
		continue
	    send_cmd(g.sock, CMD_REGISTER_FAIL, id)

    def finish_turn (self, sock, id, type):
	if self.register_turn:
	    self.register_turn = False

	self.turn.pop(0)
	if self.timeoutid:
	    gobject.source_remove(self.timeoutid)
	    self.timeoutid = None

	for i, g in self.gamers.iteritems():
	    if g.state != CLIENT_STARTED and g.state != CLIENT_REGISTERED:
		continue
	    send_cmd(g.sock, CMD_FINISH_TURN, id, type)

	if type == 1:
	    tile = self.tilestack.pop(0)
	    self.gamers[id].tiles.append(tile)
	    send_cmd(sock, CMD_GIVE_1_TILE, id, tile)
	elif type == 2:
	    self.tiles = self.copyboard
	    self.gamers[id].tiles = self.copydeck
	    tile1 = self.tilestack.pop(0)
	    tile2 = self.tilestack.pop(0)
	    tile3 = self.tilestack.pop(0)
	    self.gamers[id].tiles.append(tile1)
	    self.gamers[id].tiles.append(tile2)
	    self.gamers[id].tiles.append(tile3)
	    send_cmd(sock, CMD_GIVE_3_TILE, id, tile1, tile2, tile3)

	if not self.tilestack or not self.gamers[id].tiles:
	    self.end()
	else:
	    self.start_turn()

    def finish_turn_fail (self, sock, id):
	for i, g in self.gamers.iteritems():
	    if g.state != CLIENT_STARTED and g.state != CLIENT_REGISTERED:
		continue
	    send_cmd(g.sock, CMD_FINISH_TURN_FAIL, id)

    def check_register (self, id):
	sum = 0
	for tr, cr in zip(self.tiles, self.copyboard):
	    rowtiles = []
	    for c in range(self.drawing.board.COLS):
		if tr[c] != -1 and cr[c] == -1:
		    if tr[c] == 52 or tr[c] == 53:
			return False
		    rowtiles.append(tr[c])
		    sum += drawing.get_tile_tuple(tr[c])[1] + 1
	    if not rowtiles:
		continue
	    if not self.check_row(rowtiles, 0):
		return False
	if sum < 30:
	    return False
	return True

    def check_all_set (self, id):
	for r in self.tiles:
	    rowtiles = []
	    for t in r:
		if t != -1:
		    rowtiles.append(t)
	    if not rowtiles:
		continue
	    if 52 in rowtiles or 53 in rowtiles:
		if not self.check_joker_row(rowtiles):
		    return False
	    elif not self.check_row(rowtiles, 0):
		return False
	return True
    
    def check_joker_row (self, rowtiles):
	if 52 in rowtiles:
	    j1 = rowtiles.index(52)
	    if 53 in rowtiles:
		j2 = rowtiles.index(53)
		for i in range(52):
		    rowtiles[j1] = i
		    for j in range(52):
			rowtiles[j2] = j
			if self.check_row(rowtiles, 0):
			    return True
	    else:
		for i in range(52):
		    rowtiles[j1] = i
		    if self.check_row(rowtiles, 0):
			return True
	elif 53 in rowtiles:
	    j1 = rowtiles.index(53)
	    for i in range(52):
		rowtiles[j1] = i
		if self.check_row(rowtiles, 0):
		    return True
	return False

    def check_row (self, rowtiles, first):
	set = self.find_next_set(rowtiles, first)
	if set == None:
	    return False
	if set[1] == len(rowtiles):
	    return True
	for i in range(set[1], set[0] - 1, -1):
	    if self.check_row(rowtiles, i):
		return True
	return False
    
    def find_next_set (self, rowtiles, first):
	set = self.check_type_1(rowtiles, first)
	if set == None:
	    set = self.check_type_2(rowtiles, first)
	return set
    
    def check_type_1 (self, rowtiles, first):
	c, n = drawing.get_tile_tuple(rowtiles[first])
	end = first
	for i in range(first + 1, len(rowtiles)):
	    ci, ni = drawing.get_tile_tuple(rowtiles[i])
	    if c != ci:
		break
	    if ni != n + 1:
		break
	    n = ni
	    end = i + 1
	start = max(first + 3, end - 2)
	if start > end:
	    return None
	return start, end

    def check_type_2 (self, rowtiles, first):
	c, n = drawing.get_tile_tuple(rowtiles[first])
	color = [0] * 4
	color[c] = 1
	end = first
	for i in range(first + 1, len(rowtiles)):
	    ci, ni = drawing.get_tile_tuple(rowtiles[i])
	    if ni != n:
		break
	    if color[ci] == 1:
		break
	    color[ci] = 1
	    end = i + 1
	start = max(first + 3, end - 2)
	if start > end:
	    return None
	return start, end

    def recv_ready (self, sock, id, name):
	self.gamers[id].state = CLIENT_READY
	self.gamers[id].name = name

	for i, g in self.gamers.iteritems():
	    if g.state == CLIENT_CONNECTED:
		continue
	    send_cmd(sock, CMD_READY, i, g.name)
	    if g.sock != sock:
		send_cmd(g.sock, CMD_READY, id, name)
	
	self.number = self.number + 1
	if self.state == SERVER_CREATED and self.number >= 2:
	    self.ready()
    
    def recv_register (self, sock, id):
	if id != self.turn[0]:
	    return
	if self.gamers[id].state != CLIENT_STARTED:
	    return

	if not self.check_all_set(id):
	    self.register_fail(sock, id)
	elif not self.check_register(id):
	    self.register_fail(sock, id)
	else:
	    self.register(sock, id)

    def recv_try_finish_turn (self, sock, id):
	if id != self.turn[0]:
	    return

	valid = self.check_all_set(id)

	if not valid:
	    self.finish_turn_fail(sock, id)
	elif len(self.copydeck) == len(self.gamers[id].tiles):
	    if self.register_turn:
		self.finish_turn(sock, id, 0)
	    else:
		self.finish_turn(sock, id, 1)
	elif self.gamers[id].state == CLIENT_STARTED:
	    self.finish_turn_fail(sock, id)
	else:
	    self.finish_turn(sock, id, 0)

    def recv_move_deck_to_board (self, sock, id, tile, row, col):
	if id != self.turn[0]:
	    send_cmd(sock, CMD_MOVE_FAIL, id)
	    return
	if tile not in self.gamers[id].tiles:
	    send_cmd(sock, CMD_MOVE_FAIL, id)
	    return
	if self.tiles[row][col] != -1:
	    send_cmd(sock, CMD_MOVE_FAIL, id)
	    return
	self.gamers[id].tiles.remove(tile)
	self.tiles[row][col] = tile
	for i, g in self.gamers.iteritems():
	    if g.state != CLIENT_STARTED and g.state != CLIENT_REGISTERED:
		continue
	    send_cmd(g.sock, CMD_MOVE_DECK_TO_BOARD, id, tile, row, col)

    def recv_move_board_to_deck (self, sock, id, tile, row, col):
	if id != self.turn[0]:
	    send_cmd(sock, CMD_MOVE_FAIL, id)
	    return
	if tile not in self.copydeck:
	    send_cmd(sock, CMD_MOVE_FAIL, id)
	    return
	if self.tiles[row][col] != tile:
	    send_cmd(sock, CMD_MOVE_FAIL, id)
	    return
	self.tiles[row][col] = -1
	self.gamers[id].tiles.append(tile)
	for i, g in self.gamers.iteritems():
	    if g.state != CLIENT_STARTED and g.state != CLIENT_REGISTERED:
		continue
	    send_cmd(g.sock, CMD_MOVE_BOARD_TO_DECK, id, tile, row, col)

    def recv_move_board_to_board (self, sock, id, tile, frow, fcol, row, col):
	if self.tiles[frow][fcol] != tile:
	    send_cmd(sock, CMD_MOVE_FAIL, id)
	    return
	if self.tiles[row][col] != -1:
	    send_cmd(sock, CMD_MOVE_FAIL, id)
	    return
	self.tiles[frow][fcol] = -1
	self.tiles[row][col] = tile
	for i, g in self.gamers.iteritems():
	    if g.state != CLIENT_STARTED and g.state != CLIENT_REGISTERED:
		continue
	    send_cmd(g.sock, CMD_MOVE_BOARD_TO_BOARD, id,
		    tile, frow, fcol, row, col)

    def recv_exchange_deck_to_board (self, sock, id, ftile, ttile, row, col):
	if id != self.turn[0]:
	    send_cmd(sock, CMD_MOVE_FAIL, id)
	    return
	if ttile not in self.copydeck:
	    send_cmd(sock, CMD_MOVE_FAIL, id)
	    return
	if self.tiles[row][col] != ttile:
	    send_cmd(sock, CMD_MOVE_FAIL, id)
	    return
	self.tiles[row][col] = ftile
	self.gamers[id].tiles.remove(ftile)
	self.gamers[id].tiles.append(ttile)
	for i, g in self.gamers.iteritems():
	    if g.state != CLIENT_STARTED and g.state != CLIENT_REGISTERED:
		continue
	    send_cmd(g.sock, CMD_EXCHANGE_DECK_TO_BOARD, id,
		    ftile, ttile, row, col)

    def recv_exchange_board_to_deck (self, sock, id,
	ftile, ttile, row, col):
	if id != self.turn[0]:
	    send_cmd(sock, CMD_MOVE_FAIL, id)
	    return
	if ftile not in self.copydeck:
	    send_cmd(sock, CMD_MOVE_FAIL, id)
	    return
	if self.tiles[row][col] != ftile:
	    send_cmd(sock, CMD_MOVE_FAIL, id)
	    return
	self.tiles[row][col] = ttile
	self.gamers[id].tiles.remove(ttile)
	self.gamers[id].tiles.append(ftile)
	for i, g in self.gamers.iteritems():
	    if g.state != CLIENT_STARTED and g.state != CLIENT_REGISTERED:
		continue
	    send_cmd(g.sock, CMD_EXCHANGE_BOARD_TO_DECK, id,
		    ftile, ttile, row, col)

    def recv_exchange_board_to_board (self, sock, id, ftile, ttile,
	frow, fcol, row, col):
	if self.tiles[frow][fcol] != ftile:
	    send_cmd(sock, CMD_MOVE_FAIL, id)
	    return
	if self.tiles[row][col] != ttile:
	    send_cmd(sock, CMD_MOVE_FAIL, id)
	    return
	self.tiles[frow][fcol] = ttile
	self.tiles[row][col] = ftile
	for i, g in self.gamers.iteritems():
	    if g.state != CLIENT_STARTED and g.state != CLIENT_REGISTERED:
		continue
	    send_cmd(g.sock, CMD_EXCHANGE_BOARD_TO_BOARD, id,
		    ftile, ttile, frow, fcol, row, col)

    def recv_chat (self, sock, id, chat):
	for i, g in self.gamers.iteritems():
	    send_cmd(g.sock, CMD_CHAT, id, chat)
	
    def disconnect (self, sock):
	for i, g in self.gamers.iteritems():
	    if g.sock == sock:
		gobject.source_remove(g.io)
		g.sock.close()
		self.quit(i)
		break


    def on_start_game_button_clicked (self, button):
	self.start()

    def on_end_game_button_clicked (self, button):
	self.end()

class Client:
    turn = False

    moving = False
    gamers = {}

    def __init__ (self, xml, drawing):
	self.xml = xml
	self.xml.signal_autoconnect(self)
	self.drawing = drawing

    def run (self, host, port, name):
	self.name = name

	self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	try:
	    self.sock.connect((host, port))
	except:
	    self.sock.close()
	    raise

	self.io = gobject.io_add_watch(self.sock,
		gobject.IO_IN | gobject.IO_ERR | gobject.IO_HUP, self.receive)

    def receive (self, sock, condition):
	try:
	    cmd, args = recv_cmd(sock)
	except Disconnect:
	    self.disconnect(sock)
	    return False

	if cmd == CMD_CONNECTED:
	    id, = args
	    self.recv_connect(sock, id)
	elif cmd == CMD_READY:
	    id, name = args
	    self.recv_ready(id, name)
	elif cmd == CMD_START:
	    tiles, = args
	    self.recv_start(tiles)
	elif cmd == CMD_END:
	    score, totalscore, remain = args
	    self.recv_end(score, totalscore, remain)
	elif cmd == CMD_STOP:
	    id, = args
	    self.recv_stop(id)
	elif cmd == CMD_START_TURN:
	    id, = args
	    self.recv_start_turn(id)
	elif cmd == CMD_REGISTER:
	    id, = args
	    self.recv_register(id)
	elif cmd == CMD_REGISTER_FAIL:
	    id, = args
	    self.recv_register_fail(id)
	elif cmd == CMD_FINISH_TURN:
	    id, type = args
	    self.recv_finish_turn(id, type)
	elif cmd == CMD_FINISH_TURN_FAIL:
	    id, = args
	    self.recv_finish_turn_fail(id)
	elif cmd == CMD_GIVE_1_TILE:
	    id, tile = args
	    self.recv_give_1_tile(id, tile)
	elif cmd == CMD_GIVE_3_TILE:
	    id, tile1, tile2, tile3 = args
	    self.recv_give_3_tile(id, tile1, tile2, tile3)
	elif cmd == CMD_MOVE_DECK_TO_BOARD:
	    id, tile, row, col = args
	    self.recv_move_deck_to_board(id, tile, row, col)
	elif cmd == CMD_MOVE_BOARD_TO_DECK:
	    id, tile, row, col = args
	    self.recv_move_board_to_deck(id, tile, row, col)
	elif cmd == CMD_MOVE_BOARD_TO_BOARD:
	    id, tile, frow, fcol, row, col = args
	    self.recv_move_board_to_board(id, tile, frow, fcol, row, col)
	elif cmd == CMD_EXCHANGE_DECK_TO_BOARD:
	    id, ftile, ttile, row, col = args
	    self.recv_exchange_deck_to_board(id, ftile, ttile, row, col)
	elif cmd == CMD_EXCHANGE_BOARD_TO_DECK:
	    id, ftile, ttile, row, col = args
	    self.recv_exchange_board_to_deck(id, ftile, ttile, row, col)
	elif cmd == CMD_EXCHANGE_BOARD_TO_BOARD:
	    id, ftile, ttile, frow, fcol, row, col = args
	    self.recv_exchange_board_to_board(id,
		    ftile, ttile, frow, fcol, row, col)
	elif cmd == CMD_MOVE_FAIL:
	    id, = args
	    self.recv_move_fail(id)
	elif cmd == CMD_CHAT:
	    id, chat = args
	    self.recv_chat(id, chat)
	elif cmd == CMD_QUIT:
	    id, = args
	    self.recv_quit(id)

	return True
    
    def move_deck_to_board (self, tile, frow, fcol, row, col):
	if not self.turn:
	    return
	self.moving = True
	self.moving_data = frow, fcol
	send_cmd(self.sock, CMD_MOVE_DECK_TO_BOARD, self.me.id, tile, row, col)

    def move_board_to_deck (self, tile, frow, fcol, row, col):
	if not self.turn:
	    return
	if tile not in self.sentboard:
	    return
	self.moving = True
	self.moving_data = row, col
	send_cmd(self.sock, CMD_MOVE_BOARD_TO_DECK, self.me.id,
		tile, frow, fcol)

    def move_board_to_board (self, tile, frow, fcol, row, col):
	self.moving = True
	send_cmd(self.sock, CMD_MOVE_BOARD_TO_BOARD, self.me.id,
		tile, frow, fcol, row, col)

    def exchange_deck_to_board (self, ftile, ttile, frow, fcol, row, col):
	if not self.turn:
	    return
	if ttile not in self.sentboard:
	    return
	self.moving = True
	self.moving_data = (frow, fcol)
	send_cmd(self.sock, CMD_EXCHANGE_DECK_TO_BOARD, self.me.id,
		ftile, ttile, row, col)

    def exchange_board_to_deck (self, ftile, ttile, frow, fcol, row, col):
	if not self.turn:
	    return
	if ftile not in self.sentboard:
	    return
	self.moving = True
	self.moving_data = row, col
	send_cmd(self.sock, CMD_EXCHANGE_BOARD_TO_DECK, self.me.id,
		ftile, ttile, frow, fcol)

    def exchange_board_to_board (self, ftile, ttile,
	    frow, fcol, row, col):
	self.moving = True
	send_cmd(self.sock, CMD_EXCHANGE_BOARD_TO_BOARD, self.me.id,
		ftile, ttile, frow, fcol, row, col)

    def chat (self, chat):
	send_cmd(self.sock, CMD_CHAT, self.me.id, chat)

    def recv_connect (self, sock, id):
	self.me = Gamer(id)
	self.me.state = CLIENT_CONNECTED
	self.gamers[id] = self.me

	send_cmd(sock, CMD_READY, id, self.name)
	self.drawing.show_deck()
    
    def recv_ready (self, id, name):
	if id == self.me.id:
	    self.me.state = CLIENT_READY
	else:
	    self.gamers[id] = Gamer(id)

	self.gamers[id].name = name
	self.gamers[id].totalscore = 0
	self.drawing.add_gamer(id, name, 0)

    def recv_start (self, tiles):
	self.me.tiles = tiles
	self.me.state = CLIENT_STARTED
	self.drawing.start_round(self.me.tiles)

    def recv_end (self, score, totalscore, remain):
	self.drawing.end_round(score, totalscore, remain)
	if self.me.state != CLIENT_CONNECTED:
	    self.me.state = CLIENT_READY

    def recv_stop (self, id):
	self.drawing.stop_round()
	if self.me.state != CLIENT_CONNECTED:
	    self.me.state = CLIENT_READY

    def recv_start_turn (self, id):
	self.drawing.start_turn(id, self.gamers[id].name)
	self.copyboard = self.drawing.board.get_all_tile()

	if id == self.me.id:
	    self.sentboard = []
	    if self.me.state == CLIENT_STARTED:
		self.drawing.start_register_turn()
	    elif self.me.state == CLIENT_REGISTERED:
		self.drawing.start_normal_turn()
	    self.turn = True

    def recv_register (self, id):
	self.drawing.register_success(id, self.gamers[id].name)
	self.copyboard = self.drawing.board.get_all_tile()

	if id == self.me.id:
	    self.sentboard = []
	    self.me.state = CLIENT_REGISTERED
	    self.drawing.end_register()

    def recv_register_fail (self, id):
	self.drawing.register_fail(id, self.gamers[id].name)

    def recv_finish_turn (self, id, type):
	self.drawing.finish_turn(id, self.gamers[id].name, type)
	if type == 2:
	    self.drawing.board.set_all_tiles(self.copyboard)

	if id == self.me.id:
	    if self.me.state == CLIENT_STARTED:
		self.drawing.finish_register_turn()
	    elif self.me.state == CLIENT_REGISTERED:
		self.drawing.finish_normal_turn()
	    if type == 2:
		for t in self.sentboard:
		    self.drawing.deck.add_tile(t)
	    self.turn = False

    def recv_finish_turn_fail (self, id):
	self.drawing.finish_turn_fail(id, self.gamers[id].name)

    def recv_give_1_tile (self, id, tile):
	self.drawing.deck.add_tile(tile)

    def recv_give_3_tile (self, id, tile1, tile2, tile3):
	for t in [tile1, tile2, tile3]:
	    self.drawing.deck.add_tile(t)

    def recv_move_deck_to_board (self, id, tile, row, col):
	self.drawing.board.insert_tile(tile, row, col)
	if self.me.id == id:
	    frow, fcol = self.moving_data
	    self.drawing.deck.remove_tile(frow, fcol)
	    self.sentboard.append(tile)
	    self.moving = False

    def recv_move_board_to_deck (self, id, tile, row, col):
	self.drawing.board.remove_tile(row, col)
	if self.me.id == id:
	    row, col = self.moving_data
	    self.drawing.deck.insert_tile(tile, row, col)
	    self.sentboard.remove(tile)
	    self.moving = False

    def recv_move_board_to_board (self, id, tile, frow, fcol, row, col):
	self.drawing.board.move_tile(frow, fcol, row, col)
	self.moving = False

    def recv_exchange_deck_to_board (self, id, ftile, ttile, row, col):
	self.drawing.board.remove_tile(row, col)
	self.drawing.board.insert_tile(ftile, row, col)
	if self.me.id == id:
	    frow, fcol = self.moving_data
	    self.drawing.deck.remove_tile(frow, fcol)
	    self.drawing.deck.insert_tile(ttile, frow, fcol)
	    self.sentboard.remove(ttile)
	    self.sentboard.append(ftile)
	    self.moving = False

    def recv_exchange_board_to_deck (self, id, ftile, ttile, frow, fcol):
	self.drawing.board.remove_tile(frow, fcol)
	self.drawing.board.insert_tile(ttile, frow, fcol)
	if self.me.id == id:
	    row, col = self.moving_data
	    self.drawing.deck.remove_tile(row, col)
	    self.drawing.deck.insert_tile(ftile, row, col)
	    self.sentboard.remove(ftile)
	    self.sentboard.append(ttile)
	    self.moving = False

    def recv_exchange_board_to_board (self, id, ftile, ttile,
	    frow, fcol, row, col):
	self.drawing.board.exchange_tile(frow, fcol, row, col)
	self.moving = False

    def recv_move_fail (self, id):
	self.moving = False

    def recv_chat (self, id, chat):
	self.drawing.chat(self.gamers[id].name, chat)

    def recv_quit (self, id):
	self.drawing.del_gamer(id, self.gamers[id].name)
	del self.gamers[id]

    def disconnect (self, sock):
	gobject.source_remove(self.io)
	sock.close()
	self.drawing.del_gamer(self.me.id, self.me.name)
	self.drawing.end_game()

	self.turn = False
	self.moving = False
	self.gamers = {}

	self.xml.get_widget('new_game_toolbutton').set_sensitive(True)
	self.xml.get_widget('end_game_toolbutton').set_sensitive(False)
	self.xml.get_widget('new_game').set_sensitive(True)
	self.xml.get_widget('end_game').set_sensitive(False)

    def on_send_chat_button_clicked (self, button):
	chat_entry = self.xml.get_widget('chat_entry')
	chat = chat_entry.get_text()
	if chat != '':
	    chat_entry.set_text('')
	    self.chat(chat)

    def on_chat_entry_activate (self, entry):
	chat = entry.get_text()
	if chat != '':
	    entry.set_text('')
	    self.chat(chat)
    
    def on_register_button_clicked (self, button):
	send_cmd(self.sock, CMD_REGISTER, self.me.id)

    def on_finish_turn_button_clicked (self, button):
	send_cmd(self.sock, CMD_TRY_FINISH_TURN, self.me.id)

    def on_board_button_press_event (self, widget, event):
	if event.button != 1:
	    return False
	if self.moving:
	    return True

	row, col = self.drawing.board.tile_coord(event.x, event.y)
	if self.drawing.board.selected:
	    frow, fcol = self.drawing.board.selected
	    self.drawing.board.unselect_tile()
	    ftile = self.drawing.board.get_tile(frow, fcol)
	    tile = self.drawing.board.get_tile(row, col)
	    if tile == -1:
		self.move_board_to_board(ftile, frow, fcol, row, col)
	    else:
		self.exchange_board_to_board(ftile, tile, frow, fcol, row, col)
	elif self.drawing.deck.selected:
	    frow, fcol = self.drawing.deck.selected
	    self.drawing.deck.unselect_tile()
	    ftile = self.drawing.deck.get_tile(frow, fcol)
	    tile = self.drawing.board.get_tile(row, col)
	    if tile == -1:
		self.move_deck_to_board(ftile, frow, fcol, row, col)
	    else:
		self.exchange_deck_to_board(ftile, tile, frow, fcol, row, col)
	elif self.drawing.board.get_tile(row, col) != -1:
	    self.drawing.board.select_tile(row, col)

	return True

    def on_deck_button_press_event (self, widget, event):
	if event.button != 1:
	    return False
	if self.moving:
	    return True

	row, col = self.drawing.deck.tile_coord(event.x, event.y)
	if self.drawing.deck.selected:
	    frow, fcol = self.drawing.deck.selected
	    self.drawing.deck.unselect_tile()
	    if self.drawing.deck.get_tile(row, col) == -1:
		self.drawing.deck.move_tile(frow, fcol, row, col)
	    else:
		self.drawing.deck.exchange_tile(frow, fcol, row, col)
	elif self.drawing.board.selected:
	    frow, fcol = self.drawing.board.selected
	    self.drawing.board.unselect_tile()
	    ftile = self.drawing.board.get_tile(frow, fcol)
	    tile = self.drawing.deck.get_tile(row, col)
	    if tile == -1:
		self.move_board_to_deck(ftile, frow, fcol, row, col)
	    else:
		self.exchange_board_to_deck(ftile, tile, frow, fcol, row, col)
	elif self.drawing.deck.get_tile(row, col) != -1:
	    self.drawing.deck.select_tile(row, col)

	return True

    def on_end_game_toolbutton_clicked (self, toolbutton):
	self.disconnect(self.sock)

    def on_end_game_avtivate (self, menuitem):
	self.disconnect(self.sock)
