#!/usr/bin/env python

# vim:ts=8:sts=4:sw=4
# -*- coding: utf-8 -*-

import gtk
import gtk.glade
import socket
import gettext
import os

import drawing
import network

PACKAGE = 'rummikub'
VERSION = '0.1'
LOCALEDIR = 'locale'
XML = 'rummikub.glade'

_ = gettext.gettext

class Rummikub:
    def __init__ (self):
	gettext.bindtextdomain(PACKAGE, LOCALEDIR)
	gettext.textdomain(PACKAGE)

	gtk.glade.bindtextdomain(PACKAGE, LOCALEDIR)
	gtk.glade.textdomain(PACKAGE)

	gtk.glade.set_custom_handler(self.custom_handler)
	self.xml = gtk.glade.XML(XML)
	self.xml.signal_autoconnect(self)
	self.drawing = drawing.Drawing(self.xml)
	self.server = network.Server(self.xml, self.drawing)
	self.client = network.Client(self.xml, self.drawing)

    def custom_handler (self, glade, function_name, widget_name,
	    str1, str2, int1, int2):
	if function_name == 'create_board':
	    return drawing.create_board(widget_name, str1, int1, int2)
	return None

    def run (self):
	if 'USERNAME' in os.environ:
	    name = os.environ['USERNAME']
	else:
	    name = ''

	if len(name) != 0:
	    self.xml.get_widget('name_entry').set_text(name)

	gtk.main()

    def show_message (self, dialog, message):
	message_dialog = gtk.MessageDialog(dialog, gtk.DIALOG_MODAL,
		gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, message)
	message_dialog.connect('response',
		lambda dialog, response_id: dialog.destroy())
	message_dialog.run()

    def on_new_game_dialog_response (self, dialog, response_id):
	if response_id != gtk.RESPONSE_OK:
	    dialog.hide()
	    return

	name = self.xml.get_widget('name_entry').get_text().strip()
	host = self.xml.get_widget('host_entry').get_text().strip()
	port = self.xml.get_widget('port_entry').get_text().strip()

	if len(name) == 0:
	    self.show_message(dialog, _('No name'))
	    return
	if len(port) == 0:
	    self.show_message(dialog, _('No port'))
	    return


	if self.xml.get_widget('create_button').get_active():
	    try:
		self.server.run(int(port))
		self.client.run('localhost', int(port), name)
	    except socket.error, err:
		errno, msg = err
		self.show_message(dialog, msg)
		return
	else:
	    if len(host) == 0:
		self.show_message(dialog, _('No host'))
		return
	    try:
		self.client.run(host, int(port), name)
	    except socket.error, err:
		errno, msg = err
		self.show_message(dialog, msg)
		return

	self.xml.get_widget('new_game_toolbutton').set_sensitive(False)
	self.xml.get_widget('end_game_toolbutton').set_sensitive(True)
	self.xml.get_widget('new_game').set_sensitive(False)
	self.xml.get_widget('end_game').set_sensitive(True)
	dialog.hide()

    def on_new_game_activate (self, menuitem):
	self.xml.get_widget('new_game_dialog').run()

    def on_new_game_toolbutton_clicked (self, toolbutton):
	self.xml.get_widget('new_game_dialog').run()

    def on_quit_activate (self, menuitem):
	gtk.main_quit()

    def on_about_activate (self, menuitem):
	self.xml.get_widget('aboutdialog').run()

    def on_rummikub_destroy (self, widget):
	gtk.main_quit()

    def on_name_entry_activate (self, entry):
	self.xml.get_widget('new_game_dialog').response(gtk.RESPONSE_OK)

    def on_host_entry_activate (self, entry):
	self.xml.get_widget('new_game_dialog').response(gtk.RESPONSE_OK)

    def on_port_entry_activate (self, entry):
	self.xml.get_widget('new_game_dialog').response(gtk.RESPONSE_OK)

    def on_create_button_toggled (self, togglebutton):
	if togglebutton.get_active():
	    self.xml.get_widget('host_entry').set_sensitive(False)

    def on_join_button_toggled (self, togglebutton):
	if togglebutton.get_active():
	    self.xml.get_widget('host_entry').set_sensitive(True)

def main ():
    Rummikub().run()

if __name__ == '__main__':
    main()
